#include "Progressive2D.frag"
#include "Complex.frag"

uniform float time;

const float pi = 3.141592653;

const int p = 6;
const int q = 4;

float adj = acosh(sqrt(2.0));
float hyp = acosh(sqrt(3.0));

vec2 eye = vec2(0.1, 0.1);
float eyeSize = vec2(0.05);

vec2 face = vec2(0.0, 0.0);
vec2 edge = vec2(adj, 0.0);
vec2 vertex = vec2(hyp * cos(pi / 6.0), hyp * sin(pi / 6.0));

float edist(vec2 z, vec2 w) { return length(z - w); }
float hdist(vec2 z, vec2 w) { return acosh(1.0 + 2.0 * dot(z - w, z - w) / ((1.0 - dot(z, z)) * (1.0 - dot(w, w)))); }
float h2e(float z) { return sqrt((cosh(z) - 1.0) / (cosh(z) + 1.0)); }
vec2 h2e(vec2 z) { return normalize(z) * h2e(length(z)); }
float e2h(float z) { return hdist(vec2(z, 0.0), vec2(0.0, 0.0)); }
vec2 e2h(vec2 z) { return normalize(z) * e2h(length(z)); }

vec4 moebius(vec2 z) { return vec4(z, 1.0, 0.0); }
vec2 unmoebius(vec4 z) { return cDiv(z.xy, z.zw); }
float mlength(vec4 z) { return length(z.xy) / length(z.zw); }

mat4 rotation(float a) {
  float c = cos(a);
  float s = sin(a);
  return mat4
    ( c, -s, 0.0, 0.0
    , s,  c, 0.0, 0.0
    , 0.0, 0.0, 1.0, 0.0
    , 0.0, 0.0, 0.0, 1.0
    );
}

mat4 translation(vec2 z) {
  float s = atan(z.y, z.x);
  float l = length(z);
  float e = exp(l);
  float f = e + 1.0;
  float g = e - 1.0;
  mat4 m = mat4
    ( f, 0.0, g, 0.0
    , 0.0, f, 0.0, g
    , g, 0.0, f, 0.0
    , 0.0, g, 0.0, f
    );
  if (l > 0.0) {
    return rotation(s) * m * rotation(-s);
  } else {
    return m;
  }
}

mat4 rotationAbout(vec2 z, float a) { return translation(z) * rotation(a) * translation(-z); }

vec3 circleBetween(vec2 p, vec2 q) {
  float ax = p.x;
  float ay = p.y;
  float a2 = ax * ax + ay * ay;
  float bx = q.x;
  float by = q.y;
  float b2 = bx * bx + by * by;
  float cx = ax / a2;
  float cy = ay / a2;
  float c2 = cx * cx + cy * cy;
  float d = 2.0 * (ax * (by - cy) + bx * (cy - ay) + cx * (ay - by));
  float ux = (a2*(by - cy) + b2*(cy - ay) + c2*(ay - by)) / d;
  float uy = (a2*(cx - bx) + b2*(ax - cx) + c2*(bx - ax)) / d;
  float dx = ux - q.x;
  float dy = uy - q.y;
  float r = sqrt(dx * dx + dy * dy);
  return vec3(ux, uy, r);
}

float pi3 = pi/3.0;
mat2 m3 = mat2(cos(pi3), sin(pi3), -sin(pi3), cos(pi3));
float pi23 = 2.0*pi/3.0;
mat2 m23 = mat2(cos(pi23), sin(pi23), -sin(pi23), cos(pi23));
vec3 circleA = circleBetween(h2e(edge), m3 * h2e(edge));
vec3 circleB = circleBetween(h2e(vertex), m23 * h2e(vertex));

vec4 fish(vec2 w) {
  w = mat2(cos(pi/6.0), -sin(pi/6.0), sin(pi/6.0), cos(pi/6.0)) * w;
  mat2 r = transpose(m23);
  float t = mod(atan(w.y, w.x) * 3.0 / (2.0 * pi), 3.0);
  int k = 3;
  if (t >= 1.0) {
    w = r * w;
    k--;
  }
  if (t >= 2.0) {
    w = r * w;
    k--;
  }
  vec4 c = vec4(1.0, 0.0, 0.0, 1.0);
  float h = 0.02;
  float da1 = length(w - circleA.xy) - circleA.z;
  float da2 = length(w - m3 * circleA.xy) - circleA.z;
  float db1 = length(w - circleB.xy) - circleB.z;
  float db2 = length(w - transpose(m3) * circleB.xy) - circleB.z;
  if (db1 > -h && da1 < h) {
     c.w *= abs(min(abs(db1), abs(-da1)));
     if (db1 > 0.0 && da1 < 0.0) {
       k = (k +2) % 3;
     }
  }
  if (db2 > -h && da2 < h) {
     c.w *= abs(min(abs(db2),abs( -da2)));
     if (db2 > 0.0 && da2 < 0.0) {
       k = (k + 1) % 3;
    }
  }
  if (db1 < h && da1 < h) {
     c.w *= abs(-db1);
  }
  if (db2 < h && da2 < h) {
     c.w *= abs(-db2);
  }
  if (db1 > -h && da1 > -h) {
     c.w *= abs(da1);
  }
  if (db2 > -h && da2 > -h) {
     c.w *= abs(da2);
  }
  if (da2 < h && da1 < h) {
    c.w *= abs(min(da1, da2));
  }
  if (abs(w.y) < h) {
    c.w *= abs(w.y);
  }
  if (length(w - h2e(vertex)) < 2.2 * h) {
     c.w *= length(w - h2e(vertex)) / 2.2;
  }
  if (length(w - h2e(m3 * vertex)) < 2.2 * h) {
     c.w *= length(w - h2e(m3 * vertex)) / 2.2;
  }
  w = transpose(m23) * w;
  if (abs(w.y) < h) {
    c.w *= abs(w.y);
  }
  c.w = abs(c.w);
  c.w = clamp(128.0 * c.w, 0.0, 1.0);
  if (k == 2) { c.rgb = c.brg; }
  if (k == 1) { c.rgb = c.gbr; }
  return c;
}
// hyperbolic area D(r) = 4 pi sinh^2(r/2)
// angle deficit = - area
// pi = 4 pi sinh^2 (r / 2)
// r = 2 asinh (1/2)
// 2.0 pi / 3 = ... => r = 2 asinh(sqrt(1/6))
vec3 color(vec2 z) {
  vec4 q = moebius(z);
  vec2 p = vec2(2.0 * asinh(sqrt(1.0/120.0)), 0.0);
  float a = 2.0 * pi * time - pi / 2.0;
  float b = pi  / 30.0 * time - pi / 6.0;
  p = mat2(cos(a), -sin(a), sin(a), cos(a)) * p;
  q = translation(p) * q;
  q = rotation(-b) * q;
  q = translation(vertex) * q;
  z = unmoebius(q);
  mat4 vtransform = rotationAbout(vertex, pi / 2.0);
  mat3 vpermutation = mat3
    ( -1.0,0.0,0.0
    , 0.0,0.0,1.0
    , 0.0,-1.0,0.0
    );
  mat4 faceT= rotationAbout(face, 2.0 * pi / 3.0);
  mat3 fpermutation = mat3
    ( 0.0,1.0,0.0
    , 0.0,0.0,1.0
    , 1.0,0.0,0.0
    );
  mat4 transformation[9];
  mat3 permutation[9];
  bool parities[9];
  mat4 transform = mat4
    ( 1.0,0.0,0.0,0.0
    , 0.0,1.0,0.0,0.0
    , 0.0,0.0,1.0,0.0
    , 0.0,0.0,0.0,1.0
    );
  mat3 permute = mat3
    ( 1.0,0.0,0.0
    , 0.0,1.0,0.0
    , 0.0,0.0,1.0
    );
  int k = 0;
  for (int i = 0; i < 3; ++i) {
    mat4 t = transform;
    mat3 p = permute;
    bool par = false;
    for (int j = 0; j < 3; ++j) {
      t = vtransform * t;
      p = p * vpermutation;
      par = !par;
      transformation[k] = t;
      permutation[k] = p;
      parities[k] = par;
      ++k;
    }
    transform = faceT * transform;
    permute = permute * fpermutation;
  }
  mat3 rpermute = mat3(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);
  bool parity = false;
  float a0 = 2.0 * pi / 3.0;
  mat2 r0 = mat2(cos(a0), -sin(a0), sin(a0), cos(a0));
  mat2 r1 = mat2(1.0,0.0,0.0,1.0);
  int kk = int(mod(floor(time/5.0), 3.0));
  for (int i = 0; i < kk && i < 3; ++i) {
    r1 *= r0;
  }
  vec4 m = moebius(z);
  float height = 0.0;
  vec3 normal = vec3(0.0, 0.0, 1.0);
  if (mlength(m) < 1.0) {
    height = 8.0 * sqrt(1.0 - mlength(m) * mlength(m));
    normal = normalize(vec3(unmoebius(m), height));
    for (int i = 0; i < 6; ++i) {
      if (mlength(m) <= h2e(adj)) { break; }
      vec4 mm = m;
      mat3 rppermute = rpermute;
      bool pparity = parity;
      bool done = true;
      for (int j = 0; j < 9; ++j) {
        vec4 mmm = transformation[j] * m;
        if (mlength(mmm) < mlength(mm)) {
          mm = mmm;
          rppermute = rpermute * permutation[j];
          pparity = (parity == parities[j]);
          done = false;
        }
      }
      m = normalize(mm);
      rpermute = rppermute;
      parity = pparity;
      if (done) { break; }
    }
    float a = 2.0 * pi / 3.0;
    mat2 r = mat2(cos(a), -sin(a), sin(a), cos(a));
    vec2 w = unmoebius(m);
  }

  vec3 c = vec3(0.0);
  vec2 w = unmoebius(m);
  float v = 1.0;
  if (length(z) < 1.0) {
    float t = atan(w.y, w.x) * 6.0 / (4.0 * pi);
    float a = - pi/6.0;
    if (parity) { t = 1.5 - t; }
    mat2 r = mat2(cos(a), -sin(a), sin(a), cos(a));
    int k = int(floor(mod(t, float(p)/2.0)));
    float s = mod(atan(w.y, w.x), pi / 3.0) - pi / 6.0;
    float g = h2e(atanh(tanh(adj) / cos(s))) * cos(s) / cos(pi/6.0);
    w = r * w;
    vec4 f = fish(w);
    c = rpermute * f.rgb;
    float h = sqrt(0.75);
    mat3 cm = mat3(h, 0.5, 0.0, 0.0, h, 0.5, 0.5, 0.0, h);
    v = c.r + c.g + c.b;
    c = v < 0.0 ? cm * abs(c) : transpose(cm) * abs(c);
    v = f.a;
  }
  return mix(vec3(1.0), c, v);
}
